require "rubygems"
require "patron"
require "nokogiri"
require "always_verify_ssl_certificates"
require "capistrano"
require "net/http"
require "net/https"
require "erubis"
require "yajl/json_gem"

class Maven
  def initialize(url)
    @url = url.gsub(/\/$/, '')
    @session = Patron::Session.new
    @session.timeout = 10
    @session.base_url = @url
  end

  def find_versions(group_id, artifact_id)
    resp = @session.get("/#{urlify(group_id)}/#{artifact_id}/maven-metadata.xml")
    if resp.status == 200
      Nokogiri::XML(resp.body).css("metadata versioning versions version").map { |e| e.text.strip }
    else
      raise "unable to find metadata for #{group_id}:#{artifact_id}"
    end
  end

  def find_release_versions(group_id, artifact_id)
    find_versions(group_id, artifact_id).select { |v| v !~ /SNAPSHOT$/ }
  end

  def find_snapshot_builds(group_id, artifact_id, version)
    resp = @session.get("/#{urlify(group_id)}/#{artifact_id}/#{version}/maven-metadata.xml")
    if resp.status == 200
      Nokogiri::XML(resp.body).css('metadata > versioning > snapshotVersions').map { |e|
        ext = e.at("extension")
        if ext && ext.text.strip == "jar"
          e.at("value").text.strip
        else
          []
        end
      }.flatten
    else
      raise "unable to find metadata for #{group_id}:#{artifact_id}:#{version}"
    end
  end

  def release_artifact_url(group_id, artifact_id, version)
    "#{@url}/#{urlify(group_id)}/#{artifact_id}/#{version}/#{artifact_id}-#{version}.jar"
  end

  def snapshot_artifact_url(group_id, artifact_id, version, build)
    "#{@url}/#{urlify(group_id)}/#{artifact_id}/#{version}/#{artifact_id}-#{build}.jar"
  end

  private

  def urlify(group_id)
    group_id.gsub('.', '/')
  end
end

Capistrano::Configuration.instance(:must_exist).load do
  
  ##############################################################################
  ################################ Initialization ##############################
  ##############################################################################
  
  logger.level = Capistrano::Logger::IMPORTANT
  default_run_options[:pty] = true
  @invoke = tasks.delete(:invoke)
  @shell = tasks.delete(:shell)

  @hosts = {}
  @data = {}
  
  ##############################################################################
  ###################################### API ###################################
  ##############################################################################
  
  # Defines the artifact's groupId.
  def group_id(name)
    set :current_group_id, name
  end

  # Defines the artifact's artifactId.
  def artifact_id(name)
    set :current_artifact_id, name
  end
  
  # Defines a host with an optional host-specific configuration block.
  def host(name, &block)
    @hosts[name] = block || lambda {}
    server name, :app
  end

  # Defines an environment which might allow snapshots and an optional
  # environment-specific configuration block.
  def environment(name, opts={}, &block)
    env_block = block || lambda {}

    desc "Deploy to the #{name} environment"
    task name.to_sym do
      set :current_environment, name
      set :allow_snapshots, opts[:allow_snapshots]
      env_block.call
      execute_task(find_task("deploy"))
    end
  end

  # Defines a Maven repository.
  def maven(url)
    if exists?(:current_repo)
      abort "You're calling #maven twice. Stop it."
    else
      set :current_repo, url
    end
  end
  
  # Defines an application with an optional owner, group, and configuration
  # block.
  def application(name, opts={}, &block)
    app_block = block || lambda {}
    set :application_name, name
    set :owner, opts[:owner] || name
    set :owner_group, opts[:group] || name
    set :healthcheck_url, opts[:healthcheck]
    namespace :deploy do
      task :configure do
        setup
        puts "> Configuring the application"
        app_block.call
      end
      
      if opts[:healthcheck]
        task :verify do
          puts("> Verifying")
          for server in roles[:app].servers
            uri = URI.parse(erb(healthcheck_url))
            req = Net::HTTP::Get.new(uri.path)
            Net::HTTP.start(uri.host, uri.port) do |http|
              resp = http.request(req)
              if resp.code.to_i == 200
                puts(">> #{server.host} is OK")
              else
                puts(">> #{server.host} is DOWN!!!")
                puts(resp.body.split("\n").map { |s| ">>> #{s}" }.join("\n"))
              end
            end
          end
        end
      else
        task :verify do
          puts("> No healthcheck configured for the application. Hope it's running.")
        end
      end
    end
  end
  
  # Defines a piece of configuration data.
  def data(name, value)
    @data[name] = value
  end
  
  # Renders a template to the given path on each server.
  def template(template, path, opts={})
    puts ">> Uploading #{template} to #{path}"
    opts = { :owner => owner, :group => owner_group, :mode => "0644" }.merge(opts)
    filename = "#{temp_dir}/#{File.basename(path)}"
    parallelized(:app) do
      output = erb(File.read("templates/#{template}"))
      put output, filename
      run "#{sudo} cp #{filename} #{path}"
      run "#{sudo} chmod #{opts[:mode]} #{path}"
      run "#{sudo} chown -R #{opts[:owner]}:#{opts[:group]} #{path}"
    end
  end
  
  # Uploads an asset to the given path on each server.
  def asset(asset, path, opts={})
    puts ">> Uploading #{asset} to #{path}"
    opts = { :owner => owner, :group => owner_group, :mode => "0644" }.merge(opts)
    filename = "#{temp_dir}/#{File.basename(path)}"
    upload(File.join("assets", asset), filename)
    run "#{sudo} cp #{filename} #{path}"
    run "#{sudo} chmod #{opts[:mode]} #{path}"
    run "#{sudo} chown -R #{opts[:owner]}:#{opts[:group]} #{path}"
  end
  
  # Uploads and adds an Upstart configuration file for the application.
  def upstart(config_file)
    puts ">> Configuring Upstart for #{application_name}"
    base = "#{application_name}.conf"
    filename = "#{temp_dir}/#{base}"
    upload(File.join("assets", config_file), filename)
    run "#{sudo} cp #{filename} /etc/init/#{base}"
    run "#{sudo} ln -sf /lib/init/upstart-job /etc/init.d/#{application_name}"
  end
  
  # Post a message to your Yammer network when you deploy.
  def yammer(token, opts={})
    after :deploy do
      set :deployed_version, exists?(:build) ? "#{version} (#{build})" : version
      opts = {:host => "www.yammer.com",
              :body => %Q{<%= application_name %> v<%= deployed_version %> was just deployed to the <%= current_environment %> environment by @<%= `whoami`.strip %>.}
      }.merge(opts)
      puts "> Sending notification to Yammer"
      opts[:body] = Erubis::Eruby.new(opts[:body]).result(binding)
      http = Net::HTTP.new(opts.delete(:host), 443)
      http.use_ssl = true
      request = Net::HTTP::Post.new("/api/v1/messages.json?access_token=#{token}")
      request.set_form_data(opts)

      response = http.request(request)
      if response.is_a? Net::HTTPSuccess
        puts ">> Sent!"
      else
        puts "!! Error sending notification: #{response.code} #{response.message}"
      end
    end
  end
  
  ##############################################################################
  #################################### Tasks ###################################
  ##############################################################################
  
  namespace :deploy do
    task :default do
      check_required_parameters!
      
      puts "> Deploying #{application_name} to #{current_environment}"

      artifact_url # queue this up before hand
      
      setup
      stage
      configure
      symlink
      restart
      verify
      cleanup
      
      puts "> Successfully #{application_name} to #{current_environment}! Congrats!"
    end
    
    task :setup, :role => :app do
      run "mkdir -p #{temp_dir}"
    end
    
    task :stage, :role => :app do
      setup
      puts "> Downloading the artifact remotely"
      run "curl --silent #{artifact_url} --output #{temp_dir}/#{artifact_filename}"
      run "#{sudo} cp #{temp_dir}/#{artifact_filename} #{directory}/#{artifact_filename}"
      run "#{sudo} chown -R #{owner}:#{owner_group} #{directory}/#{artifact_filename}"
    end
    
    task :symlink, :role => :app do
      puts "> Updating the symlink"
      run "#{sudo} ln -sf #{directory}/#{artifact_filename} #{directory}/#{application_name}#{File.extname(artifact_filename)}"
    end
    
    task :cleanup, :role => :app do
      setup
      puts "> Cleaning up"
      run "rm -rf #{temp_dir}"
    end
    
    task :restart, :role => :app, :max_hosts => 1 do
      puts "> Restarting #{application_name}"
      run "#{sudo} stop #{application_name} ; #{sudo} start #{application_name} ; sleep 10"
    end
  end
  
  ##############################################################################
  ################################# Internal ###################################
  ##############################################################################
  
  def erb(text)
    Erubis::Eruby.new(text).result(@data)
  end
  
  def parallelized(role, &block)
    $parallel_block = block
    original_data = @data.dup
    for server in roles[role].servers
      @data = original_data
      @data[:host] = server.host
      @hosts[server.host].call
      $current_host = server.host
      execute_task namespaces[:execute].tasks[:parallel]
    end
    @data = original_data
  end
  
  namespace :execute do
    task :parallel, :hosts => lambda { $current_host } do
      $parallel_block.call
    end
  end
  
  set :temp_dir do
    "/tmp/#{application_name}-#{rand(10_000_000)}"
  end
  
  set(:artifact_url) do
    hl = HighLine.new
    m = Maven.new(current_repo)

    hl.say("Select a version to deploy:")
    versions = if allow_snapshots
                 m.find_versions(current_group_id, current_artifact_id).reverse[0..2]
               else
                 m.find_release_versions(current_group_id, current_artifact_id).reverse[0..2]
               end

    set :version, hl.choose(*versions)
    puts

    if version =~ /SNAPSHOT$/
      builds = m.find_snapshot_builds(current_group_id, current_artifact_id, version)
      set :build, build = hl.ask("Which build? ") { |q| q.default = builds.first }
      puts
      m.snapshot_artifact_url(current_group_id, current_artifact_id, version, build)
    else
      m.release_artifact_url(current_group_id, current_artifact_id, version)
    end
  end

  set(:artifact_filename) do
    filename = artifact_url.split("/").last
    "#{application_name}-#{Time.now.strftime("%Y%m%d%H%M%S")}-#{version}#{File.extname(filename)}"
  end
  
  set(:directory) { "/opt/#{application_name}" }
  
  def check_required_parameters!
    unless exists?(:application_name)
      abort("No application defined.")
    end
    
    unless exists?(:current_repo)
      abort("No Maven repository defined.")
    end

    unless exists?(:current_group_id)
      abort("No group ID defined.")
    end

    unless exists?(:current_artifact_id)
      abort("No artifact ID defined.")
    end
    
    unless exists?(:current_environment)
      abort("No target environment specified. Please use one of the to:* tasks.")
    end
  end
  
  task :trace do
    logger.level = Capistrano::Logger::TRACE
    tasks[:invoke] = @invoke
    tasks[:shell] = @shell
  end
end

